"""This module provides a User class."""

from ..auth import Client
from .resource import Resource


class User(Resource):
    """A class for working with Swimlane users."""

    def __init__(self, fields):
        """Init a User with fields.

        Args:
            fields (dict): A dict of fields and values.
        """
        super(User, self).__init__(fields)

    @classmethod
    def find_all(cls):
        """List all users.

        Returns:
            A generator that yields all users in the system.
        """
        return (User(u) for u in Client.get("user"))

    @classmethod
    def find(cls, user_id=None, name=None):
        """Find users.

        Args:
            user_id (str): The user ID.
            name (str): A name or a fragment of a name to search by.

        Returns:
            A single User if user_id was specified, otherwise a generator
            that yields Users for all system users who's username matched.
        """
        if user_id:
            return User(Client.get("user/{0}".format(user_id)))

        return (User(u) for u
                in Client.get("user/search?query={0}".format(name)))



"""This module provides a Report class."""

from datetime import datetime
from ..auth import Client
from .. import SwimlaneDict
from .resource import Resource


class Report(Resource):
    """A report class used for searching."""

    def __init__(self, fields):
        """Init a Report with fields.

        Args:
            fields (dict): A dict of fields and values.
        """
        super(Report, self).__init__(fields)

    def insert(self):
        """Insert the current Report."""
        self._fields = Client.post(self, "reports")

    def update(self):
        """Update the current Report."""
        self._fields = Client.put(self, "reports")

    @classmethod
    def new_for(cls, app_id, user_id, name):
        """Get a prefilled Report for the App designated by app_id.

        Args:
            app_id (str): The ID of the App to search in.
            user_id (str): The ID of the user creating the report.
            name (str): The name of the Report.

        Return:
            A prefilled Report.
        """
        created = datetime.utcnow().isoformat() + "Z"
        return Report(SwimlaneDict({
            "$type": "Core.Models.Search.Report, Core",
            "groupBys": [],
            "aggregates": [],
            "applicationIds": [app_id],
            "columns": [],
            "sorts": {
                "$type": "System.Collections.Generic.Dictionary`2"
                         "[[System.String, mscorlib],"
                         "[Core.Models.Search.SortTypes, Core]], mscorlib",
            },
            "filters": [],
            "pageSize": 18,
            "offset": 0,
            "defaultSearchReport": False,
            "allowed": [],
            "permissions": {
                "$type": "Core.Models.Security.PermissionMatrix, Core"
            },
            "createdDate": created,
            "modifiedDate": created,
            "createdByUser": user_id,
            "modifiedByUser": user_id,
            "id": None,
            "name": name,
            "disabled": False,
            "keywords": ""
        }))

    @classmethod
    def find_all(cls, app_id=None):
        """Find all reports.

        If app_id is specified, only reports that are a member of that App
        will be returned. By default, all reports in the system are returned.
        This method will return either Reports or StatsReports depending on
        what Swimlane returns.

        Args:
            app_id (str): An App ID.

        Returns:
            A generator that yields Reports.
        """
        url = "reports/all"
        if app_id:
            url += "?appId={0}".format(app_id)

        return (Report(r) for r in Client.get(url))

    @classmethod
    def find(cls, report_id):
        """Find a report by ID.

        Args:
            report_id (str): A Report ID.

        Returns:
            A Report.
        """
        return Report(Client.get("reports/{0}".format(report_id)))




from .resource import Resource
from .app import App
from .record import Record
from .report import Report
from .stats_report import StatsReport
from .stats_result import StatsResult
from .user import User
from .group import Group
from .task import Task



"""
This module provides a username/password-based authentication provider
to be used by Client.
"""


import requests
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin


COOKIE_NAME = ".AspNet.ApplicationCookie"


class UserPassAuthProvider(object):
    """An authentication provider based on a username and password."""

    def __init__(self, base_url, username, password):
        """Init a provider that can authenticate with Swimlane using
        a username and a password.

        Args:
            base_url (str): The base URL for Swimlane
            username (str): A Swimlane username
            password (str): The password

        Returns:
            UserPassAuthProvider: An instance of a UserPassAuthProvider
        """
        self.base_url = urljoin(base_url, "user/login")
        self.username = username
        self.password = password

    def auth_header(self):
        """Get an auth header that can be used for HTTP requests.

        Returns:
            dict: A dict that can be converted to an HTTP request header.
        """
        creds = {"username": self.username, "password": self.password}
        resp = requests.post(self.base_url, data=creds)

        # Raise any underlying HTTPErrors if they occured
        resp.raise_for_status()

        return {"Cookie": COOKIE_NAME + "=" + resp.cookies[COOKIE_NAME]}



"""Provides an App class."""

from ..auth import Client
from .resource import Resource


class App(Resource):
    """A simple abstraction over a Swimlane app."""

    def __init__(self, fields):
        """Init an App with fields.

        Args:
            fields (dict): A dict of fields and values
        """
        super(App, self).__init__(fields)

    def field_id(self, name):
        """Get the field ID of a field by name.

        Args:
            name (str): The name of the field.

        Returns:
            A field ID as a str.
        """
        return next((f["id"] for f in self.fields if f["name"] == name), None)

    @classmethod
    def find_all(cls):
        """List all apps.

        Returns:
            A generator that yields all apps in the system.
        """
        return (App(x) for x in Client.get("apps/"))

    @classmethod
    def find(cls, app_id=None, name=None, acronym=None):
        """Find an application.

        Args:
            app_id (str): The app ID
            name (str): The app name
            acronym (str): The app acronym

        Returns:
            A resource that matches the fields
        """
        if app_id:
            return App(Client.get("app/{0}".format(app_id)))

        return next(
            (a for a in cls.find_all()
             if a.name == name or a.acronym == acronym), None)



"""This module provides a simple resource class."""

__metaclass__ = type


class Resource:
    """A simple abstraction over a Swimlane resource."""

    def __init__(self, fields):
        """Init a Resource with fields."""
        self._fields = fields

    def __getattr__(self, name):
        """Allow the internal fields dict to be retrievable as an attr.

        Args:
            name (str): The attr name.

        Returns:
            The value of self.fields[name] or AttributeError if name is
            not in self.fields.
        """
        if name in self._fields:
            return self._fields[name]

        raise AttributeError(name)

    def __setattr__(self, name, value):
        """Allow the internal fields dict to be settable as an attr.

        Args:
            name (str): The attr name.
            value: The value to set.
        """
        if "_fields" in self.__dict__ and name in self.__dict__["_fields"]:
            self.__dict__["_fields"][name] = value
            return

        self.__dict__[name] = value

    def __str__(self):
        """A string version."""
        return str(self._fields)



"""Helper class to make Python play nice with Swimlane."""

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


class SwimlaneDict(OrderedDict):
    """A helper class that knows about JSON.NET $type values."""

    def __init__(self, d):
        """Init a SwimlaneDict with a dict.

        The dict is recursively converted to a series of OrderDicts with all
        "$type" keys bubbled to the top. This allows a SwimlaneDict to be
        serialized to JSON in a way that is compliant with JSON.NET.

        Args:
            d (dict): An unordered dict.

        Returns:
            A SwimlaneDict
        """
        self._ordered(d)
        super(SwimlaneDict, self).__init__(sorted(d.items()))

    def _ordered(self, d):
        for k, v in d.items():
            if isinstance(v, dict):
                d[k] = OrderedDict(sorted(v.items()))
                self._ordered(d[k])
